<?php

class	stl_point {
	var	$parent = null;
	var	$id = 0;
	var	$x = 0;
	var	$y = 0;
	var	$z = 0;
	var	$children;
	function	stl_point(&$parent, $id = 0, $x = 0, $y = 0, $z = 0) {
		$this->parent =& $parent;
		$this->id = $id;
		$this->x = $x;
		$this->y = $y;
		$this->z = $z;
		$this->children = array();
	}
	function	moveto($x = 0, $y = 0, $z = 0) {
		$this->parent->beforemove($x - $this->x, $y - $this->y, $z - $this->z);
		$oldpoint =& new stl_point($this->parent, -1, $this->x, $this->y, $this->z);
		$this->x = $x;
		$this->y = $y;
		$this->z = $z;
		$this->parent->aftermove($this->id, $oldpoint);
		foreach ($this->children as $key => $dummy)
			$this->children[$key]->moveto($x, $y, $z);
	}
	function	moveadd($x = 0, $y = 0, $z = 0) {
		$this->moveto($this->x + $x, $this->y + $y, $this->z + $z);
	}
	function	tokey() {
		return "@".$this->x."@".$this->y."@".$this->z;
	}
	function	addchild() {
		$this->children[] =& $this->parent->addpoint($this->x, $this->y, $this->z);
	}
}

class	stl_solid {
	var	$pointlist;
	var	$polygonlist;
	var	$moved = 0;
	var	$closed = 0;
	var	$reverse = 0;
	function	stl_solid() {
		$this->pointlist = array();
		$this->polygonlist = array();
	}
	function	&addpoint($x = 0, $y = 0, $z = 0) {
		if (($this->moved))
			die("stl_solid::addpoint called after stl_point::moveto() called.");
		$point =& new stl_point($this, count($this->pointlist), $x, $y, $z);
		$this->pointlist[] =& $point;
		return $point;
	}
	function	output($mode = 0, $offsetx = 0, $offsety = 0, $offsetz = 0) {
		switch ($mode) {
			case	1:
				return <<<EOO
solid 

EOO;
			case	2:
			case	-1:
				return <<<EOO
endsolid 

EOO;
		}
		if ($this->moved == 0)
			die("stl_solid::output stl_point::moveto() not called.");
		if ($this->closed == 0) {
			$this->addcurrentpoint(0);
			$this->closed = 1;
		}
		$s = "";
		$format = <<<EOO
facet normal %7e %7e %7e
outer loop
vertex %7e %7e %7e
vertex %7e %7e %7e
vertex %7e %7e %7e
endloop
endfacet

EOO;
		foreach ($this->polygonlist as $key => $val) {
			if ($val == 0)
				continue;
			if ($val < 0)
				list(, $x0, $y0, $z0, $x2, $y2, $z2, $x1, $y1, $z1) = explode("@", $key);
			else
				list(, $x0, $y0, $z0, $x1, $y1, $z1, $x2, $y2, $z2) = explode("@", $key);
			$ax = $x1 - $x0;
			$ay = $y1 - $y0;
			$az = $z1 - $z0;
			$bx = $x2 - $x0;
			$by = $y2 - $y0;
			$bz = $z2 - $z0;
			$vx = $ay * $bz - $az * $by;
			$vy = $az * $bx - $ax * $bz;
			$vz = $ax * $by - $ay * $bx;
			$l = pow($vx * $vx + $vy * $vy + $vz * $vz, 0.5);
			if ($l <= 0.1)
				continue;
			$s .= sprintf($format, $vx / $l, $vy / $l, $vz / $l, $x0 + $offsetx, $y0 + $offsety, $z0 + $offsetz, $x1 + $offsetx, $y1 + $offsety, $z1 + $offsetz, $x2 + $offsetx, $y2 + $offsety, $z2 + $offsetz);
		}
		return $s;
	}
	function	beforemove() {
		if (($this->moved))
			return;
		if (count($this->pointlist) < 3)
			die("stl_solid::beforemove no enough points.");
		if (($this->closed))
			die("stl_solid::beforemove called after output().");
		$this->addcurrentpoint(1);
		$this->moved = 1;
	}
	function	aftermove($id, &$oldpoint) {
		$count = count($this->pointlist);
		$this->addpolygon($this->pointlist[($id + $count - 1) % $count], $oldpoint, $this->pointlist[$id]);
		$this->addpolygon($oldpoint, $this->pointlist[($id + 1) % $count], $this->pointlist[$id]);
	}
	function	addpolygon(&$p0, &$p1, &$p2) {
#printf("addpolygon: p0(%d, %d, %d) p1(%d, %d, %d) p2(%d, %d, %d)\n", $p0->x, $p0->y, $p0->z, $p1->x, $p1->y, $p1->z, $p2->x, $p2->y, $p2->z);
		$k0 = $p0->tokey();
		$k1 = $p1->tokey();
		$k2 = $p2->tokey();
		if (@$this->polygonlist[$s = $k0.$k1.$k2] !== null)
			$this->polygonlist[$s]++;
		else if (@$this->polygonlist[$s = $k1.$k2.$k0] !== null)
			$this->polygonlist[$s]++;
		else if (@$this->polygonlist[$s = $k2.$k0.$k1] !== null)
			$this->polygonlist[$s]++;
		else if (@$this->polygonlist[$s = $k2.$k1.$k0] !== null)
			$this->polygonlist[$s]--;
		else if (@$this->polygonlist[$s = $k0.$k2.$k1] !== null)
			$this->polygonlist[$s]--;
		else if (@$this->polygonlist[$s = $k1.$k0.$k2] !== null)
			$this->polygonlist[$s]--;
		else
			$this->polygonlist[$k0.$k1.$k2] = 1;
	}
	function	product2d($id0, $id1, $id2) {
		$p0 =& $this->pointlist[$id0];
		$p1 =& $this->pointlist[$id1];
		$p2 =& $this->pointlist[$id2];
		$ax = $p1->x - $p0->x;
		$ay = $p1->y - $p0->y;
		$bx = $p2->x - $p0->x;
		$by = $p2->y - $p0->y;
#printf("  a(%d, %d) b(%d, %d)\n", $ax, $ay, $bx, $by);
		return $ax * $by - $bx * $ay;
	}
	function	addcurrentpoint($reverse = 0) {
		$count = count($this->pointlist);
		if ($this->moved == 0) {
			$sum = 0;
			for ($i=0; $i<$count; $i++) {
				$ax = $this->pointlist[$i]->x;
				$ay = $this->pointlist[$i]->y;
				$bx = $this->pointlist[($i + 1) % $count]->x;
				$by = $this->pointlist[($i + 1) % $count]->y;
				$sum += $ax * $by - $ay * $bx;
			}
			if ($sum < 0) {
#die("reverse.");
				$reverselist = array();
				for ($i=$count-1; $i>=0; $i--) {
					$obj =& $this->pointlist[$i];
					$obj->id = count($reverselist);
					$reverselist[] =& $obj;
				}
				$this->pointlist =& $reverselist;
			}
		}
		$done = array();
		for (;;) {
			$p = array();
			$oldx = $oldy = null;
			for ($i=0; $i<$count; $i++) {
				if ((@$done[$i]))
					continue;
				if (($oldx === $this->pointlist[$i]->x)&&($oldy === $this->pointlist[$i]->y))
					continue;
				$p[] = $i;
				$oldx = $this->pointlist[$i]->x;
				$oldy = $this->pointlist[$i]->y;
			}
			if (count($p) < 3)
				break;
			$first = 1;
			for (; count($p)>=3; array_shift($p)) {
#printf("p:%d(%d, %d, %d) %d(%d, %d, %d) %d(%d, %d, %d)\n", $p[0], $this->pointlist[$p[0]]->x, $this->pointlist[$p[0]]->y, $this->pointlist[$p[0]]->z, $p[1], $this->pointlist[$p[1]]->x, $this->pointlist[$p[1]]->y, $this->pointlist[$p[1]]->z, $p[2], $this->pointlist[$p[2]]->x, $this->pointlist[$p[2]]->y, $this->pointlist[$p[2]]->z);
				if ($this->product2d($p[0], $p[1], $p[2]) <= 0)
					continue;
				$i = 0;
				for (;;) {
#printf("---- i:%d(%d, %d, %d)\n", $i, $this->pointlist[$i]->x, $this->pointlist[$i]->y, $this->pointlist[$i]->z);
					if ($p[0] == $i)
						;
					else if ($p[1] == $i)
						;
					else if ($p[2] == $i)
						;
					else if (($l0 = $this->product2d($p[0], $p[1], $i)) < 0)
						;
					else if (($l1 = $this->product2d($p[1], $p[2], $i)) < 0)
						;
					else if (($l2 = $this->product2d($p[2], $p[0], $i)) < 0)
						;
					else if (($l0 <= 0)&&($l1 <= 0))
						;
					else if (($l1 <= 0)&&($l2 <= 0))
						;
					else if (($l2 <= 0)&&($l0 <= 0))
						;
					else
						break;		# point exists inside triangle
					
					if (++$i < $count)
						continue;
					
fprintf(STDERR, "---- output(%d - %d - %d)\n", $p[0], $p[1], $p[2]);
					$first = 0;
					if (($reverse))
						$this->addpolygon($this->pointlist[$p[0]], $this->pointlist[$p[2]], $this->pointlist[$p[1]]);
					else
						$this->addpolygon($this->pointlist[$p[0]], $this->pointlist[$p[1]], $this->pointlist[$p[2]]);
					$done[$p[1]] = 1;
					break;
				}
				if ($first == 0)
					break;
			}
			if (($first))
				break;
		}
	}
}


$s0 =& new stl_solid();
print $s0->output(1);

for ($i=0; $i<5; $i++) {
	$s0 =& new stl_solid();
	
	$x = $i * 12;
	$y = 0;
	$w = (3.2 + $i * 0.1) / 2 + 0.5;
	
	$a =& $s0->addpoint($x - 5, $y - 5, 0);
	$b =& $s0->addpoint($x + 5, $y - 5, 0);
	$c =& $s0->addpoint($x + 5, $y + 5, 0);
	$d =& $s0->addpoint($x - 5, $y + 5, 0);
	$a->addchild();
	$e =& $s0->addpoint($x - $w, $y - $w, 0);
	$f =& $s0->addpoint($x - $w, $y + $w, 0);
	$g =& $s0->addpoint($x + $w, $y + $w, 0);
	$h =& $s0->addpoint($x + $w, $y - $w, 0);
	$e->addchild();
	
	$e->moveadd(0.5, 0.5, 1);
	$f->moveadd(0.5, -0.5, 1);
	$g->moveadd(-0.5, -0.5, 1);
	$h->moveadd(-0.5, 0.5, 1);
	
	$e->moveadd(0, 0, 2);
	$f->moveadd(0, 0, 2);
	$g->moveadd(0, 0, 2);
	$h->moveadd(0, 0, 2);
	
	$e->moveadd(0.5, 0.5, 1);
	$f->moveadd(0.5, -0.5, 1);
	$g->moveadd(-0.5, -0.5, 1);
	$h->moveadd(-0.5, 0.5, 1);
	
	$a->moveadd(0, 0, 5);
	$b->moveadd(0, 0, 5);
	$c->moveadd(0, 0, 5);
	$d->moveadd(0, 0, 5);
	
	print $s0->output();
}
print $s0->output(2);

?>
